==> f-irc_1.26.orig.tar.gz: fi-1.26/names.c <==
/* GPLv2 applies
 * SVN revision: $Revision: 552 $
 * (C) 2006-2013 by folkert@vanheusden.com
 */
#include <stdlib.h>
#include <ncursesw/panel.h>
#include <ncursesw/ncurses.h>
#include <string.h>
#include <unistd.h>

#include "gen.h"
#include "theme.h"
#include "term.h"
#include "buffer.h"
#include "channels.h"
#include "servers.h"
#include "irc.h"
#include "names.h"
#include "utils.h"
#include "loop.h"
#include "config.h"
#include "main.h"
#include "user.h"

int names_offset = 0, names_cursor = 0;

const char *without_nick(const char *name_complete)
{
	const char *dummy = strrchr(name_complete, '!');

	if (dummy)
		return dummy + 1;

	return name_complete;
}

void free_person(person_t *p)
{
	myfree(p -> name);
	myfree(p -> name_complete);
}

void delete_index(int server_index, int channel_index, int name_index)
{
	channel *pc = &server_list[server_index].pchannels[channel_index];
	int move_n = pc -> n_names - (name_index + 1);

	free_person(&pc -> persons[name_index]);

	memmove(&pc -> persons[name_index], &pc -> persons[name_index + 1], move_n * sizeof(person_t));

	pc -> n_names--;
}

/* change the nick for use 'name_complete' to 'nick' */
void change_nick(int server_index, int channel_index, const char *nick, const char *name_complete)
{
	int name_index = -1;
	channel *cur_channel = &server_list[server_index].pchannels[channel_index];
	const char *new_nick = nick;
	char *old_nick = NULL;

	LOG("change_nick %s enter %s %s\n", cur_channel -> channel_name, str_or_nothing(nick), str_or_nothing(name_complete));

	if (nick != NULL && get_nick_mode(nick))
		new_nick = &nick[1];

	/* find the name_complete */
	if (name_complete)
	{
		char *dummy = NULL;

		old_nick = strdup(name_complete);
		dummy = strrchr(old_nick, '!');
		if (dummy) *dummy = 0x00;

		name_index = find_name(server_index, channel_index, name_complete);

		if (name_index != -1)
		{
			delete_index(server_index, channel_index, name_index);

			if (new_nick)
			{
				add_name(server_index, channel_index, new_nick, name_complete);
			}
			else
			{
				char *dummy = strdup(name_complete);
				char *dummy2 = strrchr(dummy, '!');

				if (dummy2) *dummy2 = 0x00;

				add_name(server_index, channel_index, dummy, name_complete);

				myfree(dummy);
			}

			myfree(old_nick);

			return;
		}

		LOG("name_complete %s not found\n", name_complete);
	}

	/* find the old nick */
	if (old_nick)
	{
		name_index = find_nick(server_index, channel_index, old_nick);

		if (name_index != -1)
		{
			delete_index(server_index, channel_index, name_index);

			if (new_nick)
			{
				add_name(server_index, channel_index, new_nick, str_or_nothing(name_complete));
			}
			else
			{
				LOG("ERROR! no new nick known!\n");
			}

			LOG("change_nick: set old nick %s to new nick %s and username %s\n", old_nick, new_nick, name_complete);

			myfree(old_nick);

			return;
		}

		LOG("old_nick %s not found\n", old_nick);
	}

	/* user is not known! */
	if (new_nick)	/* ...but we know the (new) nick */
		add_name(server_index, channel_index, new_nick, name_complete);
	else
		add_name(server_index, channel_index, old_nick, name_complete);

	myfree(old_nick);

	LOG("change_nick exit\n");
}

void name_bin_search(channel *cur_channel, const char *nick, int *found_at, int *insert_at)
{
        int imin = 0, imax = cur_channel -> n_names - 1;

        while(imax >= imin)
        {
                int imid = (imin + imax) / 2;
                int cmp = strcasecmp(cur_channel -> persons[imid].name, nick);

                if (cmp < 0)
                        imin = imid + 1;
                else if (cmp > 0)
                        imax = imid - 1;
                else
                {
                        *found_at = imid;
			*insert_at = -1;
                        return;
                }
        }

	*insert_at = imin;
	*found_at = -1;
}

void add_name(int server_index, int channel_index, const char *nick, const char *name_complete)
{
	int insert_index = 0, nick_found_at = -1;
	channel *cur_channel = &server_list[server_index].pchannels[channel_index];
	char *nick_compare = NULL;
	char nick_mode = 0;
	const char *nc_wo_nick = NULL;

	LOG("add_name %s enter %s %s\n", cur_channel -> channel_name, str_or_nothing(nick), str_or_nothing(name_complete));

	/* create nick from name_complete, if neccessary */
	if (nick == NULL && name_complete == NULL)
	{
		LOG("both nick and name_complete are NULL!\n");
		return;
	}

	if (name_complete)
		nc_wo_nick = without_nick(name_complete);

	if (nick)
	{
		if (get_nick_mode(nick))
		{
			nick_mode = get_nick_mode(nick);
			nick_compare = strdup(&nick[1]);
		}
		else
		{
			nick_compare = strdup(nick);
		}
	}
	else /* if no nick is known, name_complete MUST be there */
	{
		char *dummy = NULL;

		nick_compare = strdup(name_complete);

		dummy = strrchr(nick_compare, '!');
		if (dummy)
			*dummy = 0x00;
	}

	/* find position to add */
	name_bin_search(cur_channel, nick_compare, &nick_found_at, &insert_index);

	if (nick_found_at == -1)
	{
		int new_n = cur_channel -> n_names + 1, move_n = -1, loop = -1;

		if (insert_index == -1 || insert_index > cur_channel -> n_names)
		{
			LOG("insert_index range err, %s[%s] %d,%d[%d]\n", nick_compare, cur_channel -> channel_name, nick_found_at, insert_index, cur_channel -> n_names);
		}

		/* grow arrays */
		cur_channel -> persons = realloc(cur_channel -> persons, new_n * sizeof(person_t));

		/* insert? */
		move_n = cur_channel -> n_names - insert_index;

		LOG("n: %d, insert at %d, new n %d, move_n %d\n", cur_channel -> n_names, insert_index, new_n, move_n);
		if (move_n > 0)
		{
			LOG("moving %d\n", move_n);

			memmove(&(cur_channel -> persons)[insert_index + 1], &(cur_channel -> persons)[insert_index], sizeof(person_t) * move_n);
		}

		/* set */
		(cur_channel -> persons)[insert_index].name_complete = strdup(str_or_nothing(nc_wo_nick));
		(cur_channel -> persons)[insert_index].name = strdup(nick_compare);
		(cur_channel -> persons)[insert_index].mode = nick_mode;
		(cur_channel -> persons)[insert_index].ignored = FALSE;

		for(loop=0; loop<n_ignores; loop++)
		{
			if (strcasecmp(ignores[loop].channel, cur_channel -> channel_name) == 0 &&
			    strcasecmp(ignores[loop].nick, nick_compare) == 0)
			{
				(cur_channel -> persons)[insert_index].ignored = TRUE;
				break;
			}
		}

		cur_channel -> n_names++;
	}

	myfree(nick_compare);

	LOG("add_name exit\n");
}

int find_nick(int server_index, int channel_index, const char *nick)
{
	channel *cur_channel = &server_list[server_index].pchannels[channel_index];
	const char *nick_compare = nick;
	int rc = -1;
	int left = 0, right = 0;

	LOG("find_nick %s %s\n", cur_channel -> channel_name, str_or_nothing(nick));

	if (nick != NULL && get_nick_mode(nick))
		nick_compare = &nick[1];

	/* binary search */
	left = 0;
	right = cur_channel -> n_names - 1;

	while(left <= right)
	{
		int compare = 0;
		int mid = (left + right) / 2;
		person_t *person = &cur_channel -> persons[mid];

		if (get_nick_mode(&person -> mode))
			compare = strcasecmp(&(person -> name)[1], nick_compare);
		else
			compare = strcasecmp(person -> name, nick_compare);

		if (compare == 0)
		{
			rc = mid;
			break;
		}

		if (compare > 0)
			right = mid - 1;
		else
			left = mid + 1;
	}

	LOG("find_nick exit %d\n", rc);

	return rc;
}

int find_name(int server_index, int channel_index, const char *name)
{
	int name_index = -1;
	channel *cur_channel = &server_list[server_index].pchannels[channel_index];
	int rc = -1;
	const char *nc_wo_nick = NULL;

	if (!name)
		return -1;

	nc_wo_nick = without_nick(name);

	LOG("find_name %s %s\n", cur_channel -> channel_name, str_or_nothing(name));
	LOG("# names: %d\n", cur_channel -> n_names);
	LOG("name: %s\n", nc_wo_nick);

	/* find the name_complete */
	for(name_index=0; name_index<cur_channel -> n_names; name_index++)
	{
		if (strcasecmp(cur_channel -> persons[name_index].name_complete, nc_wo_nick) == 0)
		{
			rc = name_index;
			break;
		}
	}

	LOG("find_name exit %d\n", rc);

	return rc;
}

void set_name_mode(int server_index, int channel_index, const char *nick, char mode)
{
	channel *cur_channel = &server_list[server_index].pchannels[channel_index];

	int name_index = find_nick(server_index, channel_index, nick);
	if (name_index != -1)
		(cur_channel -> persons)[name_index].mode = mode;
	else
	{
		LOG("failed to set nickmode (%d) for %s: nick not found\n", mode, nick);
	}
}

BOOL delete_name(int server_index, int channel_index, const char *nick, const char *complete_name)
{
	int name_index = -1;
	channel *cur_channel = &server_list[server_index].pchannels[channel_index];
	const char *nick_compare = nick;
	BOOL rc = FALSE;

	LOG("delete_name %s enter %s %s\n", cur_channel -> channel_name, str_or_nothing(nick), str_or_nothing(complete_name));

	if (nick_compare)
	{
		if (get_nick_mode(nick_compare))
			nick_compare = &nick[1];
	}

	/* search on nick first */
	if (nick_compare != NULL)
		name_index = find_nick(server_index, channel_index, nick_compare);
	/* then try on username */
	if (name_index == -1 && complete_name != NULL)
		name_index = find_name(server_index, channel_index, without_nick(complete_name));
	/* last try: pick nick from username */
	if (name_index == -1 && complete_name != NULL)
	{
		char *dummy = strrchr(complete_name, '!');
		if (dummy) *dummy = 0x00;
		name_index = find_nick(server_index, channel_index, complete_name);
	}
	/* if either of them are found, delete */
	if (name_index != -1)
	{
		delete_index(server_index, channel_index, name_index);

		rc = TRUE;
	}
	else
	{
		LOG("delete_name: user '%s' / '%s' not found @ %d/%d\n", str_or_nothing(nick), str_or_nothing(complete_name), server_index, channel_index);
	}

	LOG("delete_name exit %d\n", rc);

	return rc;
}

void free_names_list(channel *pc)
{
	int name_index;

	for(name_index=0; name_index<pc -> n_names; name_index++)
		free_person(&pc -> persons[name_index]);

	myfree(pc -> persons);

	pc -> persons = NULL;
	pc -> n_names = 0;
}

char convert_nick_mode(char nick_mode)
{
	if (nick_mode & MODE_OPERATOR)
		return '@';

	if (nick_mode & MODE_CAN_SPEAK)
		return '+';

	return ' ';
}

void show_name_list(int server_index, int channel_index, NEWWIN *channel_window)
{
	if (server_index == -1 || channel_index == -1)
		return;

	LOG("show_name_list %d (%d) %d (%d)\n", server_index, n_servers, channel_index, server_index < n_servers ? server_list[server_index].n_channels : -123);

	if (server_index < n_servers && server_list[server_index].n_channels > channel_index)
	{
		int name_index = 0;
		channel *cur_channel = &server_list[server_index].pchannels[channel_index];

		werase(channel_window -> win);

		for(name_index=names_offset; name_index<min(names_offset + channel_window -> nlines, cur_channel -> n_names); name_index++)
		{
			char nick_mode = (cur_channel -> persons)[name_index].mode;
			BOOL ignored   = (cur_channel -> persons)[name_index].ignored;
			char nick_char = ignored ? 'I' : convert_nick_mode(nick_mode);

			if (name_index - names_offset == names_cursor) mywattron(channel_window -> win, A_REVERSE);

			limit_print(channel_window, channel_window -> ncols, name_index - names_offset, 0, "%c%s", nick_char, (cur_channel -> persons)[name_index].name);

			if (name_index - names_offset == names_cursor) mywattroff(channel_window -> win, A_REVERSE);
		}
	}
}

char get_nick_mode(const char *nick)
{
	char mode = 0;

	if (nick[0] == '@')
		mode |= MODE_OPERATOR;

	if (nick[0] == '+')
		mode |= MODE_CAN_SPEAK;

	return mode;
}

void go_to_last_name(void)
{
	int n_names = cur_channel() -> n_names;

	if (n_names >= channel_window -> nlines)
	{
		names_cursor = channel_window -> nlines - 1;
		names_offset = n_names - channel_window -> nlines;
	}
	else
	{
		names_cursor = n_names - 1;
		names_offset = 0;
	}
}

void do_names_keypress(int c)
{
	LOG("cursorkeys names\n");
	if (c == KEY_UP)
	{
		if (names_cursor > 0)
			names_cursor--;
		else if (names_offset > 0)
			names_offset--;
		else
			go_to_last_name();
	}
	else if (c == KEY_END)
	{
		go_to_last_name();
	}
	else if (c == KEY_PPAGE)
	{
		if (names_offset >= channel_window -> nlines)
		{
			names_cursor = 0;
			names_offset -= channel_window -> nlines;
		}
		else
		{
			names_cursor = 0;
			names_offset = 0;
		}
	}
	else if (c == KEY_DOWN)
	{
		if (names_offset + names_cursor < cur_channel() -> n_names - 1)
		{
			if (names_cursor < channel_window -> nlines - 1)
				names_cursor++;
			else
				names_offset++;
		}
		else
		{
			names_cursor = 0;
			names_offset = 0;
		}
	}
	else if (c == KEY_NPAGE)
	{
		if (names_offset + channel_window -> nlines < cur_channel() -> n_names)
		{
			names_cursor = 0;
			names_offset += channel_window -> nlines;
		}
		else
		{
			names_cursor = 0;
			names_offset = cur_channel() -> n_names - 1;
		}
	}
	else if (c == KEY_HOME)
	{
		names_cursor = 0;
		names_offset = 0;
	}
	else if (c == KEY_LEFT)	/* leave nameslist */
	{
		cursor_mode = CM_CHANNELS;
	}
	else if (c == KEY_RIGHT && names_offset + names_cursor < cur_channel() -> n_names)	/* enter user menu */
	{
		if (user_menu(current_server, current_server_channel_nr, names_offset + names_cursor) == -1)
		{
			cur_server() -> state = STATE_DISCONNECTED;
			update_statusline(current_server, current_server_channel_nr, "Connection to %s:%d closed", cur_server() -> server_host, cur_server() -> server_port);
		}
	}
}

==> f-irc_1.30.orig.tar.gz: fi-1.30/names.c <==
/* GPLv2 applies
 * SVN revision: $Revision: 645 $
 * (C) 2006-2013 by folkert@vanheusden.com
 */
#include <stdlib.h>
#include <ncursesw/panel.h>
#include <ncursesw/ncurses.h>
#include <string.h>
#include <unistd.h>

#include "gen.h"
#include "theme.h"
#include "term.h"
#include "buffer.h"
#include "channels.h"
#include "servers.h"
#include "irc.h"
#include "names.h"
#include "utils.h"
#include "loop.h"
#include "config.h"
#include "main.h"
#include "user.h"
#include "ignores.h"

int names_offset = 0, names_cursor = 0;

const char *without_nick(const char *name_complete)
{
	const char *dummy = strrchr(name_complete, '!');

	if (dummy)
		return dummy + 1;

	return name_complete;
}

void free_person(person_t *p)
{
	myfree(p -> name);
	myfree(p -> name_complete);
}

void delete_index(int server_index, int channel_index, int name_index)
{
	channel *pc = &server_list[server_index].pchannels[channel_index];
	int move_n = pc -> n_names - (name_index + 1);

	free_person(&pc -> persons[name_index]);

	memmove(&pc -> persons[name_index], &pc -> persons[name_index + 1], move_n * sizeof(person_t));

	pc -> n_names--;
}

/* change the nick for use 'name_complete' to 'nick' */
void change_nick(int server_index, int channel_index, const char *nick, const char *name_complete)
{
	int name_index = -1;
	const char *new_nick = nick;
	char *old_nick = NULL;

	if (nick != NULL && get_nick_mode(nick))
		new_nick = &nick[1];

	/* find the name_complete */
	if (name_complete)
	{
		char *dummy = NULL;

		old_nick = strdup(name_complete);
		dummy = strrchr(old_nick, '!');
		if (dummy) *dummy = 0x00;

		name_index = find_name(server_index, channel_index, name_complete);

		if (name_index != -1)
		{
			delete_index(server_index, channel_index, name_index);

			if (new_nick)
			{
				add_name(server_index, channel_index, new_nick, name_complete);
			}
			else
			{
				char *dummy = strdup(name_complete);
				char *dummy2 = strrchr(dummy, '!');

				if (dummy2) *dummy2 = 0x00;

				add_name(server_index, channel_index, dummy, name_complete);

				myfree(dummy);
			}

			myfree(old_nick);

			return;
		}
	}

	/* find the old nick */
	if (old_nick)
	{
		name_index = find_nick(server_index, channel_index, old_nick);

		if (name_index != -1)
		{
			delete_index(server_index, channel_index, name_index);

			if (new_nick)
			{
				add_name(server_index, channel_index, new_nick, str_or_nothing(name_complete));
			}
			else
			{
				LOG("ERROR! no new nick known!\n");
			}

			myfree(old_nick);

			return;
		}
	}

	/* user is not known! */
	if (new_nick)	/* ...but we know the (new) nick */
		add_name(server_index, channel_index, new_nick, name_complete);
	else
		add_name(server_index, channel_index, old_nick, name_complete);

	myfree(old_nick);
}

void name_bin_search(channel *cur_channel, const char *nick, int *found_at, int *insert_at)
{
        int imin = 0, imax = cur_channel -> n_names - 1;

        while(imax >= imin)
        {
                int imid = (imin + imax) / 2;
                int cmp = strcasecmp(cur_channel -> persons[imid].name, nick);

                if (cmp < 0)
                        imin = imid + 1;
                else if (cmp > 0)
                        imax = imid - 1;
                else
                {
                        *found_at = imid;
			*insert_at = -1;
                        return;
                }
        }

	*insert_at = imin;
	*found_at = -1;
}

void add_name(int server_index, int channel_index, const char *nick, const char *name_complete)
{
	int insert_index = 0, nick_found_at = -1;
	channel *cur_channel = &server_list[server_index].pchannels[channel_index];
	char *nick_compare = NULL;
	char nick_mode = 0;
	const char *nc_wo_nick = NULL;

	/* create nick from name_complete, if neccessary */
	if (nick == NULL && name_complete == NULL)
	{
		LOG("both nick and name_complete are NULL!\n");
		return;
	}

	if (name_complete)
		nc_wo_nick = without_nick(name_complete);

	if (nick)
	{
		if (get_nick_mode(nick))
		{
			nick_mode = get_nick_mode(nick);
			nick_compare = strdup(&nick[1]);
		}
		else
		{
			nick_compare = strdup(nick);
		}
	}
	else /* if no nick is known, name_complete MUST be there */
	{
		char *dummy = NULL;

		nick_compare = strdup(name_complete);

		dummy = strrchr(nick_compare, '!');
		if (dummy)
			*dummy = 0x00;
	}

	/* find position to add */
	name_bin_search(cur_channel, nick_compare, &nick_found_at, &insert_index);

	if (nick_found_at == -1)
	{
		int new_n = cur_channel -> n_names + 1, move_n = -1;

		if (insert_index == -1 || insert_index > cur_channel -> n_names)
		{
			LOG("insert_index range err, %s[%s] %d,%d[%d]\n", nick_compare, cur_channel -> channel_name, nick_found_at, insert_index, cur_channel -> n_names);
		}

		/* grow arrays */
		cur_channel -> persons = realloc(cur_channel -> persons, new_n * sizeof(person_t));

		/* insert? */
		move_n = cur_channel -> n_names - insert_index;

		if (move_n > 0)
			memmove(&(cur_channel -> persons)[insert_index + 1], &(cur_channel -> persons)[insert_index], sizeof(person_t) * move_n);

		/* set */
		(cur_channel -> persons)[insert_index].name_complete = strdup(str_or_nothing(nc_wo_nick));
		(cur_channel -> persons)[insert_index].name = strdup(nick_compare);
		(cur_channel -> persons)[insert_index].mode = nick_mode;
		(cur_channel -> persons)[insert_index].ignored = FALSE;

		(cur_channel -> persons)[insert_index].ignored = check_ignore(cur_channel -> channel_name, nick_compare, nc_wo_nick?nc_wo_nick:IGNORE_NOT_SET);

		cur_channel -> n_names++;
	}

	myfree(nick_compare);
}

int find_nick(int server_index, int channel_index, const char *nick)
{
	channel *cur_channel = &server_list[server_index].pchannels[channel_index];
	const char *nick_compare = nick;
	int rc = -1;
	int left = 0, right = 0;

	if (nick != NULL && get_nick_mode(nick))
		nick_compare = &nick[1];

	/* binary search */
	left = 0;
	right = cur_channel -> n_names - 1;

	while(left <= right)
	{
		int compare = 0;
		int mid = (left + right) / 2;
		person_t *person = &cur_channel -> persons[mid];

		if (get_nick_mode(&person -> mode))
			compare = strcasecmp(&(person -> name)[1], nick_compare);
		else
			compare = strcasecmp(person -> name, nick_compare);

		if (compare == 0)
		{
			rc = mid;
			break;
		}

		if (compare > 0)
			right = mid - 1;
		else
			left = mid + 1;
	}

	return rc;
}

int find_name(int server_index, int channel_index, const char *name)
{
	int name_index = -1;
	channel *cur_channel = &server_list[server_index].pchannels[channel_index];
	int rc = -1;
	const char *nc_wo_nick = NULL;

	if (!name)
		return -1;

	nc_wo_nick = without_nick(name);

	/* find the name_complete */
	for(name_index=0; name_index<cur_channel -> n_names; name_index++)
	{
		if (strcasecmp(cur_channel -> persons[name_index].name_complete, nc_wo_nick) == 0)
		{
			rc = name_index;
			break;
		}
	}

	return rc;
}

void set_name_mode(int server_index, int channel_index, const char *nick, char mode)
{
	channel *cur_channel = &server_list[server_index].pchannels[channel_index];

	int name_index = find_nick(server_index, channel_index, nick);
	if (name_index != -1)
		(cur_channel -> persons)[name_index].mode = mode;
	else
	{
		LOG("failed to set nickmode (%d) for %s: nick not found\n", mode, nick);
	}
}

BOOL delete_name(int server_index, int channel_index, const char *nick, const char *complete_name)
{
	int name_index = -1;
	const char *nick_compare = nick;
	BOOL rc = FALSE;

	if (nick_compare)
	{
		if (get_nick_mode(nick_compare))
			nick_compare = &nick[1];
	}

	/* search on nick first */
	if (nick_compare != NULL)
		name_index = find_nick(server_index, channel_index, nick_compare);
	/* then try on username */
	if (name_index == -1 && complete_name != NULL)
		name_index = find_name(server_index, channel_index, without_nick(complete_name));
	/* last try: pick nick from username */
	if (name_index == -1 && complete_name != NULL)
	{
		char *dummy = strrchr(complete_name, '!');
		if (dummy) *dummy = 0x00;
		name_index = find_nick(server_index, channel_index, complete_name);
	}
	/* if either of them are found, delete */
	if (name_index != -1)
	{
		delete_index(server_index, channel_index, name_index);

		rc = TRUE;
	}
	else
	{
		LOG("delete_name: user '%s' / '%s' not found @ %d/%d\n", str_or_nothing(nick), str_or_nothing(complete_name), server_index, channel_index);
	}

	return rc;
}

void free_names_list(channel *pc)
{
	int name_index;

	for(name_index=0; name_index<pc -> n_names; name_index++)
		free_person(&pc -> persons[name_index]);

	myfree(pc -> persons);

	pc -> persons = NULL;
	pc -> n_names = 0;
}

char convert_nick_mode(char nick_mode)
{
	if (nick_mode & MODE_OPERATOR)
		return '@';

	if (nick_mode & MODE_CAN_SPEAK)
		return '+';

	return ' ';
}

void show_name_list(int server_index, int channel_index, NEWWIN *channel_window)
{
	if (server_index == -1 || channel_index == -1)
		return;

	if (server_index < n_servers && server_list[server_index].n_channels > channel_index)
	{
		int name_index = 0;
		channel *cur_channel = &server_list[server_index].pchannels[channel_index];

		werase(channel_window -> win);

		for(name_index=names_offset; name_index<min(names_offset + channel_window -> nlines, cur_channel -> n_names); name_index++)
		{
			char nick_mode = (cur_channel -> persons)[name_index].mode;
			BOOL ignored   = (cur_channel -> persons)[name_index].ignored;
			char nick_char = ignored ? 'I' : convert_nick_mode(nick_mode);

			if (name_index - names_offset == names_cursor) mywattron(channel_window -> win, A_REVERSE);

			limit_print(channel_window, channel_window -> ncols, name_index - names_offset, 0, "%c%s", nick_char, (cur_channel -> persons)[name_index].name);

			if (name_index - names_offset == names_cursor) mywattroff(channel_window -> win, A_REVERSE);
		}
	}
}

char get_nick_mode(const char *nick)
{
	char mode = 0;

	if (nick[0] == '@')
		mode |= MODE_OPERATOR;

	if (nick[0] == '+')
		mode |= MODE_CAN_SPEAK;

	return mode;
}

void go_to_last_name(void)
{
	int n_names = cur_channel() -> n_names;

	if (n_names >= channel_window -> nlines)
	{
		names_cursor = channel_window -> nlines - 1;
		names_offset = n_names - channel_window -> nlines;
	}
	else
	{
		names_cursor = n_names - 1;
		names_offset = 0;
	}
}

void do_names_keypress(int c)
{
	if (c == KEY_UP)
	{
		if (names_cursor > 0)
			names_cursor--;
		else if (names_offset > 0)
			names_offset--;
		else
			go_to_last_name();
	}
	else if (c == KEY_END)
	{
		go_to_last_name();
	}
	else if (c == KEY_PPAGE)
	{
		if (names_offset >= channel_window -> nlines)
		{
			names_cursor = 0;
			names_offset -= channel_window -> nlines;
		}
		else
		{
			names_cursor = 0;
			names_offset = 0;
		}
	}
	else if (c == KEY_DOWN)
	{
		if (names_offset + names_cursor < cur_channel() -> n_names - 1)
		{
			if (names_cursor < channel_window -> nlines - 1)
				names_cursor++;
			else
				names_offset++;
		}
		else
		{
			names_cursor = 0;
			names_offset = 0;
		}
	}
	else if (c == KEY_NPAGE)
	{
		if (names_offset + channel_window -> nlines < cur_channel() -> n_names)
		{
			names_cursor = 0;
			names_offset += channel_window -> nlines;
		}
		else
		{
			names_cursor = 0;
			names_offset = cur_channel() -> n_names - 1;
		}
	}
	else if (c == KEY_HOME)
	{
		names_cursor = 0;
		names_offset = 0;
	}
	else if (c == KEY_LEFT)	/* leave nameslist */
	{
		set_cursor_mode(CM_CHANNELS);
	}
	else if (c == KEY_RIGHT && names_offset + names_cursor < cur_channel() -> n_names)	/* enter user menu */
	{
		if (user_menu(current_server, current_server_channel_nr, names_offset + names_cursor) == -1)
		{
			cur_server() -> state = STATE_DISCONNECTED;
			close(cur_server() -> fd);
			update_statusline(current_server, current_server_channel_nr, "Connection to %s:%d closed (7)", cur_server() -> server_host, cur_server() -> server_port);
		}
	}
}
